/************************************************************
*Konstantin Dorohov fro hackaday.io
* based on https://github.com/hnguy169/RM3100-Arduino/tree/main/RM3100_Arduino_SPI 
* WHEN: 18-FEB-2025
* WHAT: ESP32+WIFI+WWW 
* Reading and sending data every 15000ms to ThingSpeak
* averaging permanently 10 window
*/
#include <SPI.h>
#include <WiFi.h>
#include "RunningAverage.h"
String s_ssid="xxxxxxx";
String s_wifi_pwd="xxxxx";
String s_api_key="xxxxx"; 
char server[] = "api.thingspeak.com";
int connection_timeout=50;
int v_avg_count(10);
long t0;

//pin definitions
#define PIN_CS 5 //magnetometer
#define PIN_DRDY 34 // magnetometer


//internal register values without the R/W bit
#define RM3100_REVID_REG 0x36 // Hexadecimal address for the Revid internal register
#define RM3100_POLL_REG 0x00 // Hexadecimal address for the Poll internal register
#define RM3100_CMM_REG 0x01 // Hexadecimal address for the Continuous Measurement Mode internal register
#define RM3100_STATUS_REG 0x34 // Hexadecimal address for the Status internal register
#define RM3100_CCX1_REG 0x04 // Hexadecimal address for the Cycle Count X1 internal register
#define RM3100_CCX0_REG 0x05 // Hexadecimal address for the Cycle Count X0 internal register
#define RM3100_TMRC_REG 0x0B // Hexadecimal address for Sets Continuous Measurement Mode Data Rate

//options
#define initialCC 400 // Set the cycle count

RunningAverage myRA(v_avg_count);
float gain;
double uT, uT_avg, uT_avg_prev, uT_avg_delta;
uint8_t x2, x1, x0, v_y2, v_y1, v_y0, z2, z1, z0;
long x,y,z;
uint8_t revid;
uint16_t cycleCount;


void connectWiFi(){
    int timeout_counter = 0;
    WiFi.mode(WIFI_STA); //Optional
    WiFi.begin(s_ssid, s_wifi_pwd);
    while (WiFi.status() != WL_CONNECTED){
        Serial.print(".");
        delay(200);
        timeout_counter++;
        if(timeout_counter >= connection_timeout){
        Serial.println("Restarting");
        ESP.restart();
        }
    }
    Serial.println("");
    Serial.print("Local ESP32 IP: ");
    Serial.println(WiFi.localIP());
    }

void httpRequest(double fieldData) {
  WiFiClient client;
  if (!client.connect(server, 80)){
      Serial.println("Connection failed");
      client.stop();
      return;     
    }
   String data = "field1=" + String(fieldData);     
   client.println("POST /update HTTP/1.1");
   client.println("Host: api.thingspeak.com");
   client.println("Connection: close");
   client.println("User-Agent: ESP32WiFi/1.1");
   client.println("X-THINGSPEAKAPIKEY: "+s_api_key);
   client.println("Content-Type: application/x-www-form-urlencoded");
   client.print("Content-Length: ");
   client.print(data.length());
   client.print("\n\n");
   client.print(data);
   delay(250);
   client.stop();
}    

void setup() {
  pinMode(PIN_CS, OUTPUT);
  pinMode(PIN_DRDY, INPUT);
  uT_avg=0;
  myRA.clear();  
  digitalWrite(PIN_CS, HIGH);
  SPI.begin(); // Initiate the SPI library
  Serial.begin(115200); 
  delay(100);
 s_ssid.trim();
 s_wifi_pwd.trim();
 connectWiFi();
 
  revid = readReg(RM3100_REVID_REG);
  
  Serial.print("REVID ID = 0x"); //REVID ID should be 0x22
  Serial.println(revid, HEX);
  
  changeCycleCount(initialCC); //change the cycle count; default = 200 (lower cycle count = higher data rates but lower resolution)

  cycleCount = readReg(RM3100_CCX1_REG);
  cycleCount = (cycleCount << 8) | readReg(RM3100_CCX0_REG);

  Serial.print("Cycle Counts = "); //display cycle count
  Serial.println(cycleCount);

  gain = (0.3671 * (float)cycleCount) + 1.5; //linear equation to calculate the gain from cycle count

    // Enable transmission to take continuous measurement with Alarm functions off
  writeReg(RM3100_CMM_REG, 0x79);
  t0=millis();
   }

void loop() {
if (millis()-t0>15000) {    
   if (digitalRead(PIN_DRDY)) {
    x = 0; y = 0; z = 0;
    //read measurements
    digitalWrite(PIN_CS, LOW);
    delay(100);
    SPI.transfer(0xA4); //sends 'read' command to register 4
    x2 = SPI.transfer(0); //following 9 bytes of data
    x1 = SPI.transfer(0);
    x0 = SPI.transfer(0);
    v_y2 = SPI.transfer(0);
    v_y1 = SPI.transfer(0);
    v_y0 = SPI.transfer(0);
    z2 = SPI.transfer(0);
    z1 = SPI.transfer(0);
    z0 = SPI.transfer(0); 
    digitalWrite(PIN_CS, HIGH);
    //taking care of possible negative value
    if (x2 & 0x80) x = 0xFF;
    if (v_y2 & 0x80) y = 0xFF;
    if (z2 & 0x80) z = 0xFF;
    //  //format results into single 32 bit signed value
    x = (x * 256 * 256 * 256) | (int32_t)(x2) * 256 * 256 | (uint16_t)(x1) * 256 | x0;
    y = (y * 256 * 256 * 256) | (int32_t)(v_y2) * 256 * 256 | (uint16_t)(v_y1) * 256 | v_y0;
    z = (z * 256 * 256 * 256) | (int32_t)(z2) * 256 * 256 | (uint16_t)(z1) * 256 | z0;
    uT = sqrt(pow(((float)(x) / gain), 2) + pow(((float)(y) / gain), 2) + pow(((float)(z) / gain), 2));
    myRA.addValue(uT);  
    uT_avg = myRA.getAverage();  
       Serial.print(myRA.getAverage(),2);
      Serial.print(" "); 
      Serial.println(myRA.getStandardDeviation(),2);
      if (WiFi.status() != WL_CONNECTED) connectWiFi();                    
      httpRequest(uT_avg);
      t0=millis();
      }
    }
}       

//addr is the 7 bit value of the register's address (without the R/W bit)
uint8_t readReg(uint8_t addr){
  uint8_t data = 0;
  digitalWrite(PIN_CS, LOW);
  delay(100);
  SPI.transfer(addr | 0x80); //OR with 0x80 to make first bit(read/write bit) high for read
  data = SPI.transfer(0);
  digitalWrite(PIN_CS, HIGH);
  return data;
}

//addr is the 7 bit (No r/w bit) value of the internal register's address, data is 8 bit data being written
void writeReg(uint8_t addr, uint8_t data){
  digitalWrite(PIN_CS, LOW); 
  delay(100);
  SPI.transfer(addr & 0x7F); //AND with 0x7F to make first bit(read/write bit) low for write
  SPI.transfer(data);
  digitalWrite(PIN_CS, HIGH);
}

void changeCycleCount(uint16_t newCC){
  uint8_t CCMSB = (newCC & 0xFF00) >> 8; //get the most significant byte
  uint8_t CCLSB = newCC & 0xFF; //get the least significant byte 
  digitalWrite(PIN_CS, LOW); 
  delay(100);
  SPI.transfer(RM3100_CCX1_REG & 0x7F); //AND with 0x7F to make first bit(read/write bit) low for write
  SPI.transfer(CCMSB);  //write new cycle count to ccx1
  SPI.transfer(CCLSB);  //write new cycle count to ccx0
  SPI.transfer(CCMSB);  //write new cycle count to ccy1
  SPI.transfer(CCLSB);  //write new cycle count to ccy0
  SPI.transfer(CCMSB);  //write new cycle count to ccz1
  SPI.transfer(CCLSB);  //write new cycle count to ccz0
  digitalWrite(PIN_CS, HIGH);
}
